/* Entrypoint for node-game-outputs. v0.2.0b
 Created by nugarp/spazzy.
 
 This software is designed
 to take game memory or Mamehook (windows) as inputs and 
 generate some type of feedback response. Having the entire
 ecosystem in javascript allows us to be a lot more flexible
 than just coding in mamehooker's scripting language.
 Originally, the code only dealt with serial port
 communications for Gun4IR. Additions have been since done to allow
 for plugins that communicate with a REST endpoint, which is how we
 actually post race position information via the 'webdevice' device.
 The point is that YOU can customize it to whatever YOU need.
 Need network calls? Use webdevice for reference and go nuts in your
 version of your device.js.

 If you want to redirect mamehooker outputs to here, you will need
 the companion exe that listens to process messages and sends them
 via socket.io (which is what this app supports for mamehooker
 output communications). That companion exe is called socket-interop,
 and it will redirect mamehooker outputs in real-time to a web server
 that we run under the hood of this app.

 Changelog:
 - v0.2.0b  Code/software is licensed by CC-BY-NC-SA-4.0
 - v0.2.0  Added support for "web" deviceType. Used for leader lamp outputs for
           racing games, for example. Gun4IR uses COM deviceType.
 - v0.1.7. Added support for "quitAfterUnhook"...which should be valuable for those
           people who are running NGO just before launching a game instead of being kept
           open indefinitely.
 - v0.1.6. Added support for external calls to overrides for connected()/disconnecting()
 - v0.1.4. Changed loop interval time from 10ms to 8ms.
 - v0.1.1. Fixed a bug where a game that read memory address (TC5, HOTD4)
 would crash if mamehooker outputs were sent. Actually, they should have been
 ignored.

 Last Updated 3/3/24. */

// bunch of dependencies
const fs = require('fs');
const memoryjs = require('memoryjs');
const _ = require('lodash');
const NanoTimer = require('nanotimer');
const { performance } = require('perf_hooks');
const ph = require('./core/processHelpers');

// DEVICES
const device = require('./devices/device'); // device outputs to COM ports => Good for Gun4IR
const webdevice = require('./devices/webdevice'); // webdevice outputs to a REST target => Good for Racing Leader Lights

const memoryHelpers = require('./core/memoryHelpers')
const io = require('socket.io')();
const emitter = require('./core/emitter');

let appRefreshInterval = 1000;
let mainloopInterval = 8;
let players = [];
let openConnections = 0;

// hello world
console.log("node-game-outputs, created by nugarp/spazzy.");

// load the config
console.log("Step 1. Loading CFG");
let cfg = JSON.parse(fs.readFileSync("config/config.json"));

// update main loop interval to allow it to be overwritten by the cfg file
// for ms >= 25, nanotimer reverts to a standard timer, which may help significantly
// reduce cpu cycle-hogging when sub-ms accuracy is not needed (i.e. lamp outputs instead of solenoids.)
mainloopInterval = cfg.interval || mainloopInterval;

// listen for device connection/disconnecetion events
emitter.eventBus.on('connected', () => {
    openConnections++;
    console.log('[connected] open connections: ' + openConnections);
});
emitter.eventBus.on('disconnected', () => {
    openConnections--;
    console.log('[disconnected] open connections: ' + openConnections);
    if (openConnections === 0 && cfg.quitAfterUnhook) {
        console.log('[quitAfterUnhook] all connections closed. Hope you enjoyed!');
        process.exit();
    }
})

// detect and store which device type should be used -- COM or Web (REST)
let deviceType = "COM";
if (!cfg.deviceType || cfg.deviceType.toLowerCase() === "com") {
    deviceType = "COM";
} else if (cfg.deviceType.toLowerCase() === "web") {
    deviceType = "Web";
}

// assign the players such that the device are for [p1, p2, p3, etc.]
for (i = 0; i < cfg.players.length; i++) {
    let path = cfg.players[i];
    if (!cfg.deviceType || cfg.deviceType.toLowerCase() === "com") { // GUN4IR COM interface
        players.push(new device.Device(path, i+1, cfg.rumbleOnHook, cfg.delay, cfg.preconnect, cfg.quitAfterUnhook));
    } else if (cfg.deviceType.toLowerCase() === "web") { // Web/REST (Leader Lights) interface
        players.push(new webdevice.WebDevice(path, i+1, cfg.rumbleOnHook, cfg.delay, cfg.preconnect, cfg.quitAfterUnhook));
    }
}

// update that cfg is loaded
console.log("CFG loaded, " + deviceType + " devices set up.");

let currentApp = null;
let plugin = null;

// loop that checks for a process in memory
let checkForApps = () => {
    ph.findGames(cfg.apps).then(result => {
        if (result !== null) {
            currentApp = result;
            console.log(result);
            players.forEach(player => player.onHook(result.cfg.config.gameDelay || 0, result.cfg.key));
            let pluginLoaded = require('./plugins/' + result.cfg.plugin);
            plugin = pluginLoaded.generate(result.target.process, players, new memoryHelpers.MemoryHelper(result.target.process), result.cfg.config || {});
            players.forEach(player => player.setOnConnected(plugin._universalConnected, plugin));
            players.forEach(player => player.setOnDisconnecting(plugin._universalDisconnecting, plugin));
            plugin.hook();
        }
    });
};

let checkForAppExit = () => {
    if (currentApp === null || currentApp.rom) {  return; } // if it doesn't exist or if we are using hooked roms, ignore
    let procs = memoryjs.getProcesses();
    if (!_.find(procs, p => p.th32ProcessID === currentApp.target.th32ProcessID)) {
        console.log('UNHOOKING / APP EXIT')
        players.forEach(player => player.onUnhook());
        plugin.unhook();
        plugin = null;
        currentApp = null;
    }
}

let last_tick = performance.now();
let tick = null;


// if a plugin is loaded..call its tick method.
let mainLoop = () => {
    if (plugin !== null) {
        tick = performance.now();
        plugin.tick();
        last_tick = tick;
    }
}

// a loop to check if we need to check to find an app or if an app was closed.
let checkLoop = () => {
    if (currentApp === null) {
        checkForApps();
    } else if (currentApp !== null) {
        checkForAppExit();
    }
}

// we use a nanotimer for mainloopInterval since it is much more accurate as a timer.
tickTimer = new NanoTimer();
tickTimer.setInterval(mainLoop, '', mainloopInterval + 'm');

// get the party started...chck for apps. set up the inaccurate, standard timer, and run it now.
let apploop = setInterval(checkLoop, appRefreshInterval);
checkLoop();

// for mamehooker outputs, we are going to listen on a socket.io server
console.log('socketio server started');
io.on('connection', client => {
    console.log('socketio connected');
    client.on('OnMAMEStart', (data) => {
        // unfortunately, if an app is force exited, we might never have gotten the 
        // OnMAMEStop message, so let's assume that if we are getting an OnMAMEStart
        // we should handle that case as well and close off any potentially open prior
        // plugins.
        
        // check if we have a currentApp. if it has property ".rom", it means it came in via mamehooker.
        if (currentApp !== null && currentApp.rom && currentApp.rom.length > 0) {
            players.forEach(player => player.onUnhook());
            plugin.unhook();
            plugin = null;
            currentApp = null;
        }

        // now we can handle the OnMAMEStart aspect
        console.log('OnMAMEStart',data.Name);
        if (data && data.Name && data.Name.length > 0) {
            if (cfg.hooks && cfg.hooks.length > 0) {
                let cfgAppEntry = _.find(cfg.hooks, (h) => h.rom.toLowerCase() == data.Name.toLowerCase());
                if (cfgAppEntry) {
                    currentApp = { rom: data.Name, cfg: cfgAppEntry };
                    players.forEach(player => player.onHook(currentApp.cfg.config.gameDelay || 0, cfgAppEntry.key || currentApp.rom)); // back up to currentApp.rom if no key was specified.
                    let pluginLoaded = require('./plugins/' + currentApp.cfg.plugin);
                    plugin = pluginLoaded.generate(currentApp, players, null);
                    players.forEach(player => player.setOnConnected(plugin._universalConnected, plugin))
                    players.forEach(player => player.setOnDisconnecting(plugin._universalDisconnecting, plugin));
                    plugin.hook();
                }
            }
        }
    });

    // ONMameStop clear out the running plugin (if applicable)
    client.on('OnMAMEStop', (data) => {
        console.log('OnMAMEStop');
        if (currentApp !== null && currentApp.rom && currentApp.rom.length > 0) {
            players.forEach(player => player.onUnhook());
            plugin.unhook();
            plugin = null;
            currentApp = null;
        }
    });

    // Forward the output to the HookPlugin base class so it can update the outputs' states.
    client.on('OnMAMEOutput', (data) => {
        // memory variants don't have onMameOutput methods, hook plugins do.
        // we may get OnMameOutputs from games that we prefer to use our own memreading methods for...
        // in which case, we want to ignore the onMameOutput coming from elsewhere.
        // THAT said, maybe in the future, we might want to consider what happens to "combination" plugins
        // that read memory AND use mamehooker outputs. That would reduce additional logic that is needed.
        if (plugin !== null && plugin.onMameOutput) {
            plugin.onMameOutput(data);
        }
    });

    // Maybe we should unload the plugin if the client disconnects...
    client.on('disconnect', () => {
        console.log('socketio disconnect');
        if (currentApp !== null && currentApp.rom && currentApp.rom.length > 0) {
            players.forEach(player => player.onUnhook());
            plugin.unhook();
            plugin = null;
            currentApp = null;
        }
    });
});

// start the socketio server
io.listen(9000);