/* Feedback plugin for Custom Demulshooter outputs, v0.1.
 Please note that this file does not use genuine outputs, only
 CtmRecoil and Damaged. These don't follow real gun timings necessarily.
 But this is a generic plugin to help you get off the ground quickly...or
 for those folks who are less picky. Created by nugarp/spazzy.
 You may consider running DemulShooter with the flag
 -noinput if you are using something else for the inputs (e.g. TP).

 Last Updated 12/5/22.
 */
 
const { HookPlugin } = require('./hookplugin')

// HookPlugins can leverage the STATE object in ticks:  
// state["Property"] = { value, oldValue, changed }
// state.changed = boolean indicating ANY change since last poll

// this should work reasonably well for "fake" outputs: ctmRecoil and damaged.
// i.e. games that don't have real, genuine outputs.
// one example is Rabbids Hollywood Arcade

class DemulshooterCustomPlugin extends HookPlugin {

    #autofireHold = 39;
    #autofirePause = 38;
    #lastRecoilTime = [new Date(), new Date(), new Date(), new Date()]; //recoil is unlikely to be within 30ms of a prior fire, so we use this to ensure we don't duplicate
        // between ammo changes and PX_GunRecoil

    constructor(info, players, nil) {
        super(info, players, ["P1_CtmRecoil","P2_CtmRecoil","P3_CtmRecoil","P4_CtmRecoil","P1_Damaged","P2_Damaged","P3_Damaged","P4_Damaged"]) // register relevant outputs
        console.log(`Demulshooter-custom (${this.rom}): loaded`)
    }
    
    hook () {
        console.log(`Demulshooter-custom (${this.rom}): hooked`);
        // we'll default to ~66ms, which is roughly every 4 frames. Not all guns will likely handle shorter periods.
        // for context, hotd4/opghost is ~50ms, tranformers/let's go island are ~83ms, gunblade/la machineguns are 120ms
        // ideally, in the config.json file, these values want to be set as high possible for the fastest gun in each game.
        // ie if the fastest gun shoots every 5 frames, set this up for 5*16.67 = 83ms.
        // note that these values are also used for single shots, so if you need a long period, it may be better to have
        // a shorter autofireHold (30-80ms) and a longer autofirePause (30-Xms) where (X-afh-6) is 1 total cycle.
        this.#autofireHold = this.cfg.autofireHold || 22; // 22 + 22 + 6 = 50ms or every ~ 3 frames
        this.#autofirePause = this.cfg.autofirePause || 22;
        for (let i = 0; i < this.players.length; i++) {
            this.players[i].setAutofireTimings(this.#autofireHold, this.#autofirePause);
        }
    }

    tick () {
        if (!this.state.changed) {
            return;
        }

        let recoils = ["P1_CtmRecoil", "P2_CtmRecoil", "P3_CtmRecoil", "P4_CtmRecoil"];
        let damageds = ["P1_Damaged", "P2_Damaged", "P3_Damaged", "P4_Damaged"];
        
        for (let i = 0; i < recoils.length; i++) {
            let now = new Date();
            let allowRecoil = (now - this.#lastRecoilTime[i]) > 15; // don't duplicate recoil shots within 30ms
            let firedRecoil = !allowRecoil; // we don't want to fire it twice..recoil based on either PX_GunRecoil or PX_Ammo
            // note that PX_Ammo is not accurate, but this is a generic file anyway, as if a gun changes, PX_Ammo change too.
            // we set firedRecoil to !allowRecoil to shortcut any logic we have for setting recoil later on.
            // console.log(allowRecoil);

            if (this.players[i]) {
                // recoil
                if (this.state[recoils[i]].changed && allowRecoil) {
                    if (this.state[recoils[i]].value === 1) {
                        this.players[i].fireRecoil(1);
                        this.players[i].setRumbleEffect("GUN-RECOIL", 255, 100); // originally strong
                        this.players[i].setRumbleEffect("GUN-RECOIL2", 175, 125); // ...then fade it out
                        this.players[i].setRumbleEffect("GUN-RECOIL3", 100, 150); // ...most faded
                        firedRecoil = true;
                        this.#lastRecoilTime[i] = now;
                    }
                }

                // damage
                if (this.state[damageds[i]].changed) {
                    if (this.state[damageds[i]].value === 1) {
                        this.players[i].setRumbleEffect("HIT", 255, 350);
                    }
                }
            }
        }

        super.tick(); // sets all state.changed & state.XXX.changed variables to false.
    }

    unhook () {
        console.log(`Demulshooter-generic (${this.rom}): unhooked`);
    }
}

exports.generate = (info, players, nil) => { return new DemulshooterCustomPlugin(info, players, nil); };