/* Feedback plugin for Generic Demulshooter outputs, v0.1.7.
 Please note that this file does not use CtmRecoil, as that
 would not follow real gun timings. Created by nugarp/spazzy.
 You may consider running DemulShooter with the flag
 -noinput if you are using something else for the inputs (e.g. TP).

 v0.1.7: Add support for 4 players
 v0.1.2: Do not assume that there are two players.
 v0.1.1: Fixed lastRecoilTime to depend on player.
 Last Updated 12/6/22.
 */
 
const { HookPlugin } = require('./hookplugin')

// HookPlugins can leverage the STATE object in ticks:  
// state["Property"] = { value, oldValue, changed }
// state.changed = boolean indicating ANY change since last poll

// this should work reasonably well for:

// real gun outputs:
// tc5 (Life => Damage, Recoil => Recoil) *not tested, use native tc5 handler*
// og (Life => Damage, Recoil => Recoil) *not tested, use outputs blaster variant*
// sgg (Life => Damage, Recoil => Recoil) *not tested*

// motor gun outputs:
// gunblade (Life => Damage, Motor => Recoil) *not tested*

// derived gun outputs:
// vcop3 (Life => Damage, Ammo => Recoil) *not tested*
// confmis (Life => Damage, Ammo => Recoil) *not tested*
// hotd (Life => Damage, Ammo => Recoil) *not tested*
// vcop2 (Life => Damage, Ammo => Recoil) *not tested*
// hodsd (Life => Damage, Ammo => Recoil) *not tested*

// other games with P1_GunRecoil: (not necessary that they will work correctly though)

// WWS
// ninjasltj (DemulJvs)
// GVR Aliens
// GVR FFL
// 2 spicy
// Rambo
// LGI
// LGI3D
// Op Ghost
// SGG
// TTX Haunted Mansion 2
// TTX SHA

// other games with P1_Gunmotor: (not necessary that they will work correctly though..this code assumes they are const on, not pulsed.)

// Hikaru Braveff
// GVR FFL
// Hotd4
// LGJ
// LGJSP
// M2 BEL
// M2 Gunblade
// LGI
// LGI3D
// SDR
// Transformers
// TTX EADP
// TTX GaiaAttack4
// TTX Gundam
// TTX Haunted Museum 1 & 2

// this will NOT work well for:
// Luigi's Mansion Arcade (GunMotor & GunRecoil are both defined)..we need to separate them out for good feedback.
//   LMA may require us to read from the memory directly to determine if the vacuum is running. I can't imagine we
//   would want solenoid recoil for the game at all, but will revisit at a later time.

class DemulshooterGenericPlugin extends HookPlugin {

    #autofireHold = 39;
    #autofirePause = 38;
    #lastRecoilTime = [new Date(), new Date(), new Date(), new Date()]; //recoil is unlikely to be within 30ms of a prior fire, so we use this to ensure we don't duplicate
        // between ammo changes and PX_GunRecoil

    constructor(info, players, nil) {
        super(info, players, ["P1_GunMotor","P1_GunRecoil","P2_GunMotor","P2_GunRecoil","P3_GunMotor","P3_GunRecoil","P4_GunMotor","P4_GunRecoil",
                              "P1_Ammo","P2_Ammo","P3_Ammo","P4_Ammo","P1_Damaged","P2_Damaged","P3_Damaged","P4_Damaged"]) // register relevant outputs
        console.log(`Demulshooter-generic (${this.rom}): loaded`)
    }
    
    hook () {
        console.log(`Demulshooter-generic (${this.rom}): hooked`);
        // we'll default to ~66ms, which is roughly every 4 frames. Not all guns will likely handle shorter periods.
        // for context, hotd4/opghost is ~50ms, tranformers/let's go island are ~83ms, gunblade/la machineguns are 120ms
        // ideally, in the config.json file, these values want to be set as high possible for the fastest gun in each game.
        // ie if the fastest gun shoots every 5 frames, set this up for 5*16.67 = 83ms.
        // note that these values are also used for single shots, so if you need a long period, it may be better to have
        // a shorter autofireHold (30-80ms) and a longer autofirePause (30-Xms) where (X-afh-6) is 1 total cycle.
        this.#autofireHold = this.cfg.autofireHold || 22; // 22 + 22 + 6 = 50ms or every ~ 3 frames
        this.#autofirePause = this.cfg.autofirePause || 22;
        for (let i = 0; i < this.players.length; i++) {
            this.players[i].setAutofireTimings(this.#autofireHold, this.#autofirePause);
        }
    }

    tick () {
        if (!this.state.changed) {
            return;
        }

        let motors = ["P1_GunMotor", "P2_GunMotor", "P3_GunMotor", "P4_GunMotor"];
        let recoils = ["P1_GunRecoil", "P2_GunRecoil", "P3_GunRecoil", "P4_GunRecoil"];
        let damageds = ["P1_Damaged", "P2_Damaged", "P3_Damaged", "P4_Damaged"];
        let ammos = ["P1_Ammo", "P2_Ammo", "P3_Ammo", "P4_Ammo"];
        
        for (let i = 0; i < motors.length; i++) {
            if (this.players[i]) {
                let now = new Date();
                let allowRecoil = (now - this.#lastRecoilTime[i]) > 15; // don't duplicate recoil shots within 30ms
                let firedRecoil = !allowRecoil; // we don't want to fire it twice..recoil based on either PX_GunRecoil or PX_Ammo
                // note that PX_Ammo is not accurate, but this is a generic file anyway, as if a gun changes, PX_Ammo change too.
                // we set firedRecoil to !allowRecoil to shortcut any logic we have for setting recoil later on.
                // console.log(allowRecoil);
    
                // motors
                if (this.state[motors[i]].changed) {
                    // for autofire games we'll set the rumble a bit less to differentiate shooting a gun from the getting hit.
                    if (this.state[motors[i]].value === 1 && allowRecoil) {
                        this.players[i].startAutofire();
                        this.players[i].setRumbleEffect("GUN-FIRST", 200, 150); // make the first shot a little longer in case it was a single shot
                        this.players[i].startRumblePulsingEffect("AUTOFIRE", /* strength */ 175, /* on period */ this.#autofireHold + 3, /* off period */ this.#autofirePause + 3);
                        firedRecoil = true;
                        this.#lastRecoilTime[i] = now;
                    } else {
                        // we dont immediately stop rumble for GUN-FIRST because we want at least one shot to rumble.
                        this.players[i].stopAutofire();
                        this.players[i].stopRumblePulsingEffect("AUTOFIRE");
                    }
                }
    
                // recoil
                if (this.state[recoils[i]].changed && allowRecoil) {
                    if (this.state[recoils[i]].value === 1) {
                        this.players[i].fireRecoil(1);
                        this.players[i].setRumbleEffect("GUN-RECOIL", 255, 100); // originally strong
                        this.players[i].setRumbleEffect("GUN-RECOIL2", 175, 125); // ...then fade it out
                        this.players[i].setRumbleEffect("GUN-RECOIL3", 100, 150); // ...most faded
                        firedRecoil = true;
                        this.#lastRecoilTime[i] = now;
                    }
                }
    
                // damage
                if (this.state[damageds[i]].changed) {
                    if (this.state[damageds[i]].value === 1) {
                        this.players[i].setRumbleEffect("HIT", 255, 350);
                    }
                }
    
                // ammo
                if (this.state[ammos[i]].changed) {
                    // reload (or possibly change weapon)
                    // obviously this case will not work in the event the weapon you are changing to has -1 ammo from the last weapon
                    // but this is a generic file meant to work "well" 95% of the time, so we'll go ahead with this logic.
                    //
                    // besides, having reload and change weapon rumbles be the same seems fine to me -- they are just short "blip" rumbles.
    
                    if (this.state[ammos[i]].value != this.state[ammos[i]].oldValue) {
                        console.log("ammos changed for p" + (i+1));
                        // if (this.state[ammos[i]].value > this.state[ammos[i]].oldValue) {
                        if (this.state[ammos[i]].value !== this.state[ammos[i]].oldValue - 1) {
                            console.log("branch 1 for p" + (i+1));
                            this.players[i].setRumbleEffect("RELOAD-OR-GUN-CHANGE", 255, 100);
                        // fire (in case we didn't receive a PX_GunRecoil or PX_GunMotor and we saw ammo go down...fire the solenoid).
                        } else if (this.state[ammos[i]].value == (this.state[ammos[i]].oldValue - 1) && !firedRecoil) {
                            console.log("branch 2 for p" + (i+1));
                            this.players[i].fireRecoil(1);
                            this.players[i].setRumbleEffect("GUN-RECOIL", 255, 100); // originally strong
                            this.players[i].setRumbleEffect("GUN-RECOIL2", 175, 125); // ...then fade it out
                            this.players[i].setRumbleEffect("GUN-RECOIL3", 100, 150); // ...most faded
                            firedRecoil = true; // we don't really need to set this here but I put it for robustness if we check it later in the code.
                            this.#lastRecoilTime[i] = now;
                        }
                    }
                }    
            }
        }

        super.tick(); // sets all state.changed & state.XXX.changed variables to false.
    }

    unhook () {
        console.log(`Demulshooter-generic (${this.rom}): unhooked`);
    }
}

exports.generate = (info, players, nil) => { return new DemulshooterGenericPlugin(info, players, nil); };