/* Base class for all plugins that will utilize
 MAMEHooker outputs, v0.1.8.  Created by nugarp/spazzy.
 It will take MAMEHooker outputs and create a state object
 with all of the values that can be read during the extending
 plugins' tick() methods. Also remember to hook() and unhook()
 as needed.

 You can see tha (transformers) and lamachin (la machineguns)
 as example plugins that leverage this style of output generation.

 v0.1.8 - Add support for remapping players. May be helpful if reassigning p1 p2 as p2 p3 in 4p games.
 v0.1.7 - Add support for generic connected/disconnected files (handles multiple games at once)
 v0.1.6 - Added support for calling external files for connected() and disconnecting()
 v0.1.5 - Added support for disconnecting() callback, which is called just
 before a serial port connection is terminated. It can be called N times -
 once per player.
 v0.1.2 - Added support for connected() callback, which is called after
 a serial port connection success has been made (or if it already is connected,
 it will execute immediately). It can be called N times - once per player.
 v0.1.1 - added storage of cfg (from config.json) and rom name to
 class.

 Last Updated 2/1/23.
 */

const vm = require('vm');
const fs = require('fs');
const _ = require('lodash');
class HookPlugin {
    // info = null;
    players = null;
    state = {};
    cfg = {};
    rom = '';

    constructor (info, players, hooks = []) {
        this.rom = info.rom;
        this.process = process;
        this.cfg = info.cfg.config || {}; // get the config{} object in plugins.js

        // allow remapping of player order if needed (might be useful for 4p games)
        this.players = [];
        if (this.cfg.players) {
            this.cfg.players.forEach((port, index) => {
                this.players[index] = _.find(players, cp => cp.getInfo().port === port);
            })
        } else {
            this.players = players;
        }
        
        hooks.forEach((hook) => {
            this.state[hook] = { value: 0, oldValue: 0, changed: 0 }
        })
        this.state.changed = false;
    }

    onMameOutput (data) {
        if (this.state[data.Name] === undefined) {
            return;
        }
        console.log(data);
        let oldValue = this.state[data.Name].value;
        this.state[data.Name].value = data.State;
        this.state[data.Name].oldValue = oldValue;
        this.state[data.Name].changed = (data.State !== oldValue);
        this.state.changed = this.state.changed || this.state[data.Name].changed;
    }

    tick () {
        this.state.changed = false;
        let keys = Object.keys(this.state);
        keys.forEach((key) => {
            if (key !== 'changed') {
                this.state[key].changed = false;
            }
        });
    }

    connected (player, data) {
        console.log('connected() called:', JSON.stringify(data));
    }
    disconnecting (player, data) {
        console.log('disconnecting() called:', JSON.stringify(data));
    }
    hook () { console.log('hook() not implemented'); }
    unhook () { console.log('unhook() not implemented'); }

    // DO NOT override these _-prefix functions in your plugins
    _universalConnected (player, data) {
        this.connected(player, data);

        // if the user has defined a generic connected file, call that.
        let execFile = `${__dirname}/../overrides/connected.js`;
        if (fs.existsSync(execFile)) {
            let code = fs.readFileSync(execFile);
            vm.runInNewContext(code, {player: player, data: data});
        }
        
        // then, if they have anything even more specific for a game, call that.
        execFile = `${__dirname}/../overrides/connected/${data.key}.js`;
        if (fs.existsSync(execFile)) {
            let code = fs.readFileSync(execFile);
            vm.runInNewContext(code, {player: player, data: data, console: console});
        }
    }

    // DO NOT override these _-prefix functions in your plugins
    _universalDisconnecting (player, data) {
        this.disconnecting(player, data);

        // if the user has defined a generic disconnecting file, call that.
        let execFile = `${__dirname}/../overrides/disconnecting.js`;
        if (fs.existsSync(execFile)) {
            let code = fs.readFileSync(execFile);
            vm.runInNewContext(code, {player: player, data: data});
        }

        // then, if they have anything even more specific for a game, call that.
        execFile = `${__dirname}/../overrides/disconnecting/${data.key}.js`;
        if (fs.existsSync(execFile)) {
            let code = fs.readFileSync(execFile);
            vm.runInNewContext(code, {player: player, data: data, console: console});
        }
    }
}

module.exports = { HookPlugin: HookPlugin }