/* Feedback plugin for Jurassic Park Arcade, v0.1b.
 Created by nugarp/spazzy. Last Updated 3/1/24.

 0.1b - Bugfix because the tool would sometimes look
 for a non-populated address even after the app
 exited and load invalid values. This could lead to
 an exception. I've added better validity checks.

 0.1a - Bugfix because plugin was not unhooking when
 the game was quit due to an infinite loop issue. */

const { Plugin } = require('./plugin');
const { TYPE } = require('../core/memoryHelpers');

WEAPONS = {
    DEFAULT: 0,
    SHOCKBOLT: 4,
    TRIPLETHREAT_SHOTGUN: 6,
    FROSTBITE: 7,
    TITAN_MINIGUN: 8,
}
WEAPON_TIMINGS = {
    0: [40, 54],
    4: [0, 150],
    6: [85, 259],
    7: [0, 150],
    8: [23, 23]
}

WEAPON_INTERVALS = {
    0: 100,
    4: 200,
    6: 350,
    7: 200,
    8: 50
}

class JpaPlugin extends Plugin {
    #lastState = { players: [
        { life: -1, weapon: 0, firing: false, grenades: 0, valid: false },
        { life: -1, weapon: 0, firing: false, grenades: 0, valid: false }
    ], changed: false, valid: true }
    
    constructor(proc, players, mem, cfg) {
        console.log('JPA: loaded');
        super(proc, players, mem, cfg);
    }

    hook () {
        console.log('JPA: hooked');
        this.players[0].setAutofireTimings(40,54); // 40 + 54 + 6 = 100ms
        this.players[0].sendAction("M8.0"); // turn off full auto -- needed beacuse we will be doing S1 partially
        if (this.players[1]) {
            this.players[1].setAutofireTimings(40,54); // 40 + 54 + 6 = 100ms
            this.players[1].sendAction("M8.0"); // turn off full auto -- needed beacuse we will be doing S1 partially    
        }
    }

    unhook () {
        this.players[0].stopAutofire();
        this.players[0].sendAction("M8.1"); //auto mode
        if (this.players[1]) {
            this.players[1].stopAutofire();
            this.players[1].sendAction("M8.1"); //auto mode    
        }
        console.log("JPA: unhooked");
    }

    #isAutofire = { players: [false, false] }

    tick () {
        let state = this.#readMemory();

        // we'll treat individual players as valid instead of the entire state.
        if (!state.valid) { return; }

        let numPlayers = Math.min(this.players.length, 2)
        for (let playerId = 0; playerId < numPlayers; playerId++) {
            let playerState = state.players[playerId]
            let lastPlayerState = this.#lastState.players[playerId]

            if (playerState.valid) {
                // cases to start autofiring:
                // firing = true AND isAutofire = 0 AND player is valid
                if (!this.#isAutofire.players[playerId] && playerState.firing) {
                    console.log('start AUTOFIRE')
                    this.#isAutofire.players[playerId] = true
                    let weapon = playerState.weapon
                    let timings = WEAPON_TIMINGS[weapon]
                    let interval = WEAPON_INTERVALS[weapon]
                    this.players[playerId].setAutofireTimings(timings[0], timings[1])
                    
                    if (weapon !== WEAPONS.SHOCKBOLT && weapon !== WEAPONS.FROSTBITE) {
                        this.players[playerId].startAutofire()
                    }
                    if (weapon === WEAPONS.TRIPLETHREAT_SHOTGUN) {
                        // big rumble for the shotgun
                        this.players[playerId].startRumblePulsingEffect("AUTOFIRE", 255, 225, 75);
                        this.players[playerId].setRumbleEffect("AUTOFIRE-SHOTGUN-1", 255, 150); // in case it gets cut off early.
                        this.players[playerId].setRumbleEffect("AUTOFIRE-SHOTGUN-2", 100, 200); // in case it gets cut off early.
                        // this.players[playerId].setRumbleEffect("AUTOFIRE-SHOTGUN-3", 125, 225); // in case it gets cut off early.
                        // this.players[playerId].setRumbleEffect("AUTOFIRE-SHOTGUN-4", 75, 250); // in case it gets cut off early.
                    } else if (weapon === WEAPONS.TITAN_MINIGUN) {
                        this.players[playerId].startRumblePulsingEffect("AUTOFIRE", 255, 30, 20);
                    } else if (weapon === WEAPONS.SHOCKBOLT) {
                        this.players[playerId].startRumblePulsingEffect("AUTOFIRE", 240, 75, interval - 75);
                        this.players[playerId].startRumblePulsingEffect("AUTOFIRE-2", 150, 100, interval - 100);
                        this.players[playerId].startRumblePulsingEffect("AUTOFIRE-3", 50, 135, interval - 135);
                    } else if (weapon === WEAPONS.FROSTBITE) {
                        this.players[playerId].startRumblePulsingEffect("AUTOFIRE", 125, interval, 0);
                    } else { // std gun
                        this.players[playerId].startRumblePulsingEffect("AUTOFIRE", 175, 50, interval - 50);
                    }
                }

                // add short rumble for a weapon change (upgrade only)
                if (playerState.weapon !== lastPlayerState.weapon && playerState.weapon !== WEAPONS.DEFAULT) {
                    this.players[playerId].setRumbleEffect("WEAPON-CHANGE", 255, 100)
                }

                // cases to stop autofiring: weapon changes OR firing is not true AND isAutofire = 1
                if (this.#isAutofire.players[playerId] && (!playerState.firing || playerState.weapon !== lastPlayerState.weapon)) {
                    console.log('stop autofire')
                    this.players[playerId].stopRumblePulsingEffect('AUTOFIRE', 0)
                    this.players[playerId].stopRumblePulsingEffect('AUTOFIRE-2', 0)
                    this.players[playerId].stopRumblePulsingEffect('AUTOFIRE-3', 0)
                    this.#isAutofire.players[playerId] = false
                    this.players[playerId].stopAutofire()
                }

                // grenade launched
                if (playerState.grenades === lastPlayerState.grenades - 1) {
                    this.players[playerId].setRumbleEffect("GRENADE", 255, 200);
                    this.players[playerId].setRumbleEffect("GRENADE-2", 150, 300);
                    this.players[playerId].setRumbleEffect("GRENADE-3", 75, 400);
                }

                // hit
                if (playerState.life < lastPlayerState.life) { // we don't check valid because they could have just died.
                    console.log("HIT EVENT");
                    this.players[playerId].setRumbleEffect("HIT", 255, 500);
                }

            } else if (!playerState.valid) {
                this.players[playerId].stopRumblePulsingEffect('AUTOFIRE', 0)
                this.players[playerId].stopRumblePulsingEffect('AUTOFIRE-2', 0)
                this.players[playerId].stopRumblePulsingEffect('AUTOFIRE-3', 0)
                this.#isAutofire.players[playerId] = false
                this.players[playerId].stopAutofire()
            }
        }

        if (state.changed) { 
            console.log(JSON.stringify(state))
        }

        this.#lastState = state
    }

    // BIG THANK YOU TO ARGONLEFOU FOR THIS METHOD FROM DEMULSHOOTER.
    // COMPLETE PAIN IN THE ASS THAT P1 AND P2 LIFE ADDRESSES KEPT
    // SWAPPING FROM MY POINTERS
    #getPlayerBaseAddress(playerId) {
        let buffer1 = this.mem.read(0x08D1F720, TYPE.UINT32)
        let buffer2 = 0
        let steps = 0
        if (buffer1 !== 0) {
            while(steps < 3) { // never seen steps > 2 when logging. while(true) from ALF caused plugin to get in an infinite loop here and never unhook.
                steps += 1
                buffer2 = this.mem.read(buffer1 + 0x8, TYPE.UINT32)
                if (this.mem.read(buffer2, TYPE.BYTE) === playerId) {
                    if (this.mem.read(buffer2 + 0x10, TYPE.BYTE) !== 0) {
                        break
                    }
                }
                buffer1 = this.mem.read(buffer1, TYPE.UINT32);
                if (buffer1 === 0) {
                    return 0
                }
            }
            return buffer2
        } else {
            return 0
        }
    }

    #readMemory() {
        // let p1LifeAddr = this.mem.readChain32(this.process.modBaseAddr + 0x000EF5AC, [0x570, 0x128, 0x24, 0x4, 0x8]) + 0xC;
        // let p2LifeAddr = this.mem.readChain32(this.process.modBaseAddr + 0x0010DF00, [0xF4, 0x28, 0x4, 0x4C, 0x0, 0x8]) + 0xC;
        // let p1Life = this.mem.read(p1LifeAddr, TYPE.FLOAT)
        // let p2Life = this.mem.read(p2LifeAddr, TYPE.FLOAT)
        // let p1ShotsFired = 0
        // let p2ShotsFired = 0 //this.mem.read(p2LifeAddr + 0xB0, TYPE.UINT32)

        // these are set to 1 when the player is firing during gameplay itself (not cutscenes etc)
        let p1ValidFireAddress1 = 0x08B4BCF4 // reg weapon
        let p1ValidFireAddress2 = 0x08B4CF30 // special weapon
        let p2ValidFireAddress1 = 0x08B4BD44 // reg weapon
        let p2ValidFireAddress2 = 0x08B4CF34 // special weapon

        // sometimes the above get stuck at 1 for some reason, so we'll sanity check against whether the button was ever pushed
        let p1FiringSanityCheckAddress = 0x08DBB9FC
        let p2FiringSanityCheckAddress = 0x08DBBB54

        // thanks to ArgonLeFou for his GetPlayerRecordAddr method in Demulshooter
        let p1BaseAddress = this.#getPlayerBaseAddress(0)
        let p2BaseAddress = this.#getPlayerBaseAddress(1)

        // initializing default values
        let p1Firing = false
        let p2Firing = false
        let p1IsPlaying = false
        let p2IsPlaying = false
        let p1Weapon = 0
        let p2Weapon = 0
        let p1Grenade = 0
        let p2Grenade = 0
        let p1Life = -1
        let p2Life = -1

        if (p1BaseAddress) {
            p1IsPlaying = true
            p1Life = this.mem.read(p1BaseAddress + 0xC, TYPE.FLOAT)
            // p1ShotsFired = this.mem.read(p1BaseAddress + 0xC + 0xB0, TYPE.UINT32)
            p1Firing = !!(this.mem.read(p1ValidFireAddress1, TYPE.BYTE) || this.mem.read(p1ValidFireAddress2, TYPE.BYTE)) && this.mem.read(p1FiringSanityCheckAddress, TYPE.BYTE)
            p1Weapon = this.mem.read(0x8D20560, TYPE.UINT32)
            p1Grenade = this.mem.read(0x08B4BC84, TYPE.UINT32)
        }
        if (p2BaseAddress) {
            p2IsPlaying = true
            p2Life = this.mem.read(p2BaseAddress + 0xC, TYPE.FLOAT)
            // p2ShotsFired = this.mem.read(p2BaseAddress + 0xC + 0xB0, TYPE.UINT32)
            p2Firing = !!(this.mem.read(p2ValidFireAddress1, TYPE.BYTE) || this.mem.read(p2ValidFireAddress2, TYPE.BYTE)) && this.mem.read(p2FiringSanityCheckAddress, TYPE.BYTE)
            p2Weapon = this.mem.read(0x8D20564, TYPE.UINT32)
            p2Grenade = this.mem.read(0x08B4BC90, TYPE.UINT32)
        }

        let validWeaponIds = [0, 4, 6, 7, 8]
        
        let now = { players: [
            { life: p1Life, weapon: p1Weapon, firing: p1Firing, grenades: p1Grenade, valid: p1BaseAddress && p1Life >=0 && p1Life <= 15 && validWeaponIds.filter(wid => p1Weapon === wid) && p1Grenade >= 0 && p1Grenade <= 99 },
            { life: p2Life, weapon: p2Weapon, firing: p2Firing, grenades: p2Grenade, valid: p2BaseAddress && p2Life >=0 && p2Life <= 15 && validWeaponIds.filter(wid => p2Weapon === wid) && p2Grenade >= 0 && p2Grenade <= 99}
        ], changed: false, valid: true } // changed will be set below. valid is always true since we evaluate on a per-player basis.

        for (let i = 0; i < 2; i++) {
            now.changed = now.changed || 
                now.players[i].life !== this.#lastState.players[i].life ||
                now.players[i].weapon !== this.#lastState.players[i].weapon ||
                now.players[i].firing !== this.#lastState.players[i].firing ||
                now.players[i].grenades !== this.#lastState.players[i].grenades ||
                now.players[i].valid !== this.#lastState.players[i].valid
        }

        return now
    }
}

// todo -- pass in config & md5 info.
exports.generate = (process, players, mem) => { return new JpaPlugin(process, players, mem); };