/* Feedback plugin for The Lost World (Model 3), v0.1.
 Created by nugarp/spazzy. Last Updated 10/06/22.
 */

const { Plugin } = require('./plugin');
const { TYPE } = require('../core/memoryHelpers');

class LostWsgaPlugin extends Plugin {

    // these variables are updated in hook
    #gameBaseAddr = null;

    // temporary lastState to use at initialization
    #lastState =  { players: [ { ammo: -1, life: -1, bullets: -1 }, { ammo: -1, life: -1, bullets: -1 } ], changed: false, valid: false };

    constructor(proc, players, mem, cfg) {
        console.log('lostwsga: loaded');
        super(proc, players, mem, cfg);
    }

    hook () {
        console.log('lostwsga: hooked');
        this.players[0].setAutofireTimings(30, 30); // 30 + 30 + 6 = 66ms (just under 4 frames, which is 67ms)
        this.players[1].setAutofireTimings(30, 30);
    }

    unhook () {
        console.log("lostwsga: unhooked");
    }

    tick () {
        let state = this.#readMemory();

        if (state.changed) {
            console.log(state);
            for (let i = 0; i < state.players.length; i++) {
                let device = this.players[i]; // serial device
                let player = state.players[i]; // player info (ammo, life, bullets)
                let lastPlayer = this.#lastState.players[i];

                // health-related
                if (player.life === 0 && lastPlayer.life === 1) { // died
                    device.setRumbleEffect("DIED", 255, 600);
                } else if (player.life === lastPlayer.life - 1) { // got hit
                    device.setRumbleEffect("HIT", 255, 250);
                } else if (player.life > lastPlayer.life) { // picked up a health bar or maybe started a new game
                    device.setRumbleEffect("HEALTH-UP", 255, 100); // same feedback as reload..just a short 'tick'
                }

                // shooting-related
                // shot a bullet...check either ammo ctr goes down or total shot ctr goes up
                if (player.life > 0 && (player.ammo == lastPlayer.ammo - 1 || player.bullets == lastPlayer.bullets + 1)) {
                    device.fireRecoil(1);
                    device.setRumbleEffect("GUN-RECOIL", 255, 100); // originally strong
                    device.setRumbleEffect("GUN-RECOIL2", 175, 125); // ...then fade it out
                    device.setRumbleEffect("GUN-RECOIL3", 100, 150); // ...most faded
                }

                // reload-related
                if (player.ammo > lastPlayer.ammo) { // reload or gun change
                    device.setRumbleEffect("RELOAD-OR-GUN-CHANGE", 255, 100);
                }
            }
        }

        if (state.valid) {
            this.#lastState = state;
        }
    }

    #lastPatternCheck = new Date();

    #readMemory() {
        // this is an expensive operation so we only try it if 1 second has elapsed from the last check
        if (!this.#gameBaseAddr) {
            if ((new Date() - this.#lastPatternCheck) > 1000) {
                this.#lastPatternCheck = new Date();
                // can't find a static address so we find this pattern in memory and then the valuable addresses are based off of finding this seq.
                let addr = this.mem.findPattern('39 00 00 48 A9 00 00 48 4D 03 00 48 F9 01 00 48');
                if (addr) {
                    this.#gameBaseAddr = addr;
                } else {
                    return { valid: false };
                }
            } else {
                return { valid: false };
            }
        }

        let lookups = [
            { ammo: this.#gameBaseAddr + 0x1A3680, life: this.#gameBaseAddr + 0x1A3720, bullets: this.#gameBaseAddr + 0x1A6216 },
            { ammo: this.#gameBaseAddr + 0x1A3680 + 0x4, life: this.#gameBaseAddr + 0x1A3720 + 0x5C, bullets: this.#gameBaseAddr + 0x1A6216 + 0x10 }
        ];

        // state = { players: [ { ammo, life, bullets }, { ammo, life, bullets } ], changed, valid }
        let state = {
            players: [{},{}],
            changed: false,
            valid: true
        };

        // everything can be read as a SHORT so we can pretty nicely update everything with this loop
        lookups.forEach((L, i) => { // L = lookup object (i.e. field => address), i = player index (e.g. 0 or 1)
            Object.keys(L).forEach(f => { // f = field (e.g. "ammo", "life", "bullets")
                state.players[i][f] = this.mem.read(lookups[i][f], TYPE.SHORT);
                state.changed = state.changed || (state.players[i][f] !== this.#lastState.players[i][f]);
            })
        });

        // state.valid = state.valid && state.players[0].life <= 99 && state.players[1].life <= 99; //invalid values are quite high...1326583323, 1179648
        // state.valid = state.valid && state.players[0].bullets <= 65535 && state.players[1].bullets <= 65535 // doubt someone will shoot > 16M bullets. invalid values are 1.3B
        // state.valid = state.valid && state.players[0].ammo <= 5 && state.players[1].ammo >= 0
        return state;
    }
}

// todo -- pass in config & md5 info.
exports.generate = (process, players, mem) => { return new LostWsgaPlugin(process, players, mem); };